@@ -1,7 +1,7 @@
 /*
  * ProFTPD: mod_rewrite -- a module for rewriting FTP commands
  *
- * Copyright (c) 2001-2005 TJ Saunders
+ * Copyright (c) 2001-2009 TJ Saunders
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,13 +24,13 @@
  * This is mod_rewrite, contrib software for proftpd 1.2 and above.
  * For more information contact TJ Saunders <tj@castaglia.org>.
  *
- * $Id: mod_rewrite.c,v 1.23 2005/05/06 05:24:47 castaglia Exp $
+ * $Id: mod_rewrite.c,v 1.35 2009/02/02 18:34:52 castaglia Exp $
  */
 
 #include "conf.h"
 #include "privs.h"
 
-#define MOD_REWRITE_VERSION "mod_rewrite/0.6.8"
+#define MOD_REWRITE_VERSION "mod_rewrite/0.7"
 
 /* Make sure the version of proftpd is as necessary. */
 #if PROFTPD_VERSION_NUMBER < 0x0001021001
@@ -126,6 +126,7 @@ static char rewrite_vars[REWRITE_MAX_VARS][3] = {
 static char *rewrite_argsep(char **);
 static void rewrite_closelog(void);
 static char *rewrite_expand_var(cmd_rec *, const char *, const char *);
+static char *rewrite_get_cmd_name(cmd_rec *);
 static void rewrite_log(char *format, ...);
 static unsigned char rewrite_match_cond(cmd_rec *, config_rec *);
 static void rewrite_openlog(void);
@@ -138,6 +139,7 @@ static int rewrite_read_fifo(int, char *, size_t);
 static regex_t *rewrite_regalloc(void);
 static unsigned char rewrite_regexec(const char *, regex_t *, unsigned char,
     rewrite_match_t *);
+static void rewrite_replace_cmd_arg(cmd_rec *, char *);
 static char *rewrite_subst(cmd_rec *c, char *);
 static char *rewrite_subst_backrefs(cmd_rec *, char *, rewrite_match_t *);
 static char *rewrite_subst_maps(cmd_rec *, char *);
@@ -156,69 +158,87 @@ static int rewrite_write_fifo(int, char *, size_t);
 
 static char *rewrite_expand_var(cmd_rec *cmd, const char *subst_pattern,
     const char *var) {
-  if (!strcmp(var, "%c")) {
+  if (strcmp(var, "%c") == 0) {
     REWRITE_CHECK_VAR(session.class, "%c");
     return (session.class ? session.class->cls_name : NULL);
 
-  } else if (!strcmp(var, "%F")) {
+  } else if (strcmp(var, "%F") == 0) {
+    char *cmd_name = rewrite_get_cmd_name(cmd);
 
     /* This variable is only valid for commands that operate on paths.
      * mod_log uses the session.xfer.xfer_path variable, but that is not yet
      * set at this stage in the command dispatch cycle.
      */
-    if (!strcmp(cmd->argv[0], "APPE") || !strcmp(cmd->argv[0], "RETR") ||
-        !strcmp(cmd->argv[0], "STOR") || !strcmp(cmd->argv[0], "DELE") ||
-        !strcmp(cmd->argv[0], "MKD") || !strcmp(cmd->argv[0], "MDTM") ||
-        !strcmp(cmd->argv[0], "RMD") || !strcmp(cmd->argv[0], "SIZE") ||
-        !strcmp(cmd->argv[0], "STOU") || !strcmp(cmd->argv[0], "XMKD") ||
-        !strcmp(cmd->argv[0], "XRMD")) {
+    if (strcmp(cmd_name, "APPE") == 0 ||
+        strcmp(cmd_name, "RETR") == 0 ||
+        strcmp(cmd_name, "STOR") == 0 ||
+        strcmp(cmd_name, "DELE") == 0 ||
+        strcmp(cmd_name, "MKD") == 0 ||
+        strcmp(cmd_name, "MDTM") == 0 ||
+        strcmp(cmd_name, "RMD") == 0 ||
+        strcmp(cmd_name, "SIZE") == 0 ||
+        strcmp(cmd_name, "STOU") == 0 ||
+        strcmp(cmd_name, "XMKD") == 0 ||
+        strcmp(cmd_name, "XRMD") == 0) {
       return dir_abs_path(cmd->tmp_pool, cmd->arg, FALSE);
 
+    } else if (strcasecmp(cmd_name, "SITE CHGRP") == 0 ||
+               strcasecmp(cmd_name, "SITE CHMOD") == 0) {
+      register unsigned int i;
+      char *tmp = "";
+
+      for (i = 2; i <= cmd->argc-1; i++) {
+        tmp = pstrcat(cmd->tmp_pool, tmp, *tmp ? " " : "", cmd->argv[i], NULL);
+      } 
+
+      return dir_abs_path(cmd->tmp_pool, tmp, FALSE);
+
     } else {
-      rewrite_log("rewrite_expand_var(): %%F not valid for this command");
+      rewrite_log("rewrite_expand_var(): %%F not valid for this command ('%s')",
+        cmd_name);
       return NULL;
     }
 
-  } else if (!strcmp(var, "%f")) {
+  } else if (strcmp(var, "%f") == 0) {
     REWRITE_CHECK_VAR(cmd->arg, "%f");
     return cmd->arg;
 
-  } else if (!strcmp(var, "%m")) {
-    return cmd->argv[0];
+  } else if (strcmp(var, "%m") == 0) {
+    return rewrite_get_cmd_name(cmd);
 
-  } else if (!strcmp(var, "%p")) {
+  } else if (strcmp(var, "%p") == 0) {
     char *port = pcalloc(cmd->tmp_pool, 8 * sizeof(char));
     snprintf(port, 8, "%d", main_server->ServerPort);
     port[7] = '\0';
     return port;
 
-  } else if (!strcmp(var, "%U")) {
+  } else if (strcmp(var, "%U") == 0) {
     return get_param_ptr(main_server->conf, C_USER, FALSE);
 
-  } else if (!strcmp(var, "%P")) {
+  } else if (strcmp(var, "%P") == 0) {
     char *pid = pcalloc(cmd->tmp_pool, 8 * sizeof(char));
     snprintf(pid, 8, "%u", getpid());
     pid[7] = '\0';
     return pid;
 
-  } else if (!strcmp(var, "%g")) {
+  } else if (strcmp(var, "%g") == 0) {
     REWRITE_CHECK_VAR(session.group, "%g");
     return session.group;
 
-  } else if (!strcmp(var, "%u")) {
+  } else if (strcmp(var, "%u") == 0) {
     REWRITE_CHECK_VAR(session.user, "%u");
     return session.user;
 
-  } else if (!strcmp(var, "%a")) {
+  } else if (strcmp(var, "%a") == 0) {
     return (char *) pr_netaddr_get_ipstr(session.c->remote_addr);
 
-  } else if (!strcmp(var, "%h")) {
+  } else if (strcmp(var, "%h") == 0) {
     return (char *) session.c->remote_name;
 
-  } else if (!strcmp(var, "%v")) {
+  } else if (strcmp(var, "%v") == 0) {
     return (char *) main_server->ServerName;
 
-  } else if (!strcmp(var, "%G")) {
+  } else if (strcmp(var, "%G") == 0) {
 
     if (session.groups) {
       register unsigned int i = 0;
@@ -236,7 +256,7 @@ static char *rewrite_expand_var(cmd_rec *cmd, const char *subst_pattern,
       return NULL;
     }
 
-  } else if (!strcmp(var, "%t")) {
+  } else if (strcmp(var, "%t") == 0) {
     char *timestr = pcalloc(cmd->tmp_pool, 80 * sizeof(char));
     snprintf(timestr, 80, "%lu", time(NULL));
     timestr[79] = '\0';
@@ -288,6 +308,21 @@ static char *rewrite_argsep(char **arg) {
   return res;
 }
 
+static char *rewrite_get_cmd_name(cmd_rec *cmd) {
+  if (strcmp(cmd->argv[0], C_SITE) != 0) {
+    return cmd->argv[0];
+
+  } else {
+    if (strcasecmp(cmd->argv[1], "CHGRP") == 0 ||
+        strcasecmp(cmd->argv[1], "CHMOD") == 0) {
+      return pstrcat(cmd->pool, cmd->argv[0], " ", cmd->argv[1], NULL);
+
+    } else {
+      return cmd->argv[0];
+    }
+  }
+}
+
 static unsigned int rewrite_parse_cond_flags(pool *p, const char *flags_str) {
   char *opt = NULL, *str = NULL, **opts;
   array_header *opt_list = NULL;
@@ -310,11 +345,14 @@ static unsigned int rewrite_parse_cond_flags(pool *p, const char *flags_str) {
 
   opts = opt_list->elts;
   for (i = 0; i < opt_list->nelts; i++) {
-    if (strcmp(opts[i], "nocase") == 0 || strcmp(opts[i], "NC") == 0)
+    if (strcmp(opts[i], "nocase") == 0 ||
+        strcmp(opts[i], "NC") == 0) {
       flags |= REWRITE_COND_FLAG_NOCASE;
   
-    else if (strcmp(opts[i], "ornext") == 0 || strcmp(opts[i], "OR") == 0)
+    } else if (strcmp(opts[i], "ornext") == 0 ||
+               strcmp(opts[i], "OR") == 0) {
       flags |= REWRITE_COND_FLAG_ORNEXT;
+    }
   }
 
   return flags;
@@ -342,11 +380,14 @@ static unsigned int rewrite_parse_rule_flags(pool *p, const char *flags_str) {
 
   opts = opt_list->elts;
   for (i = 0; i < opt_list->nelts; i++) {
-    if (strcmp(opts[i], "nocase") == 0 || strcmp(opts[i], "NC") == 0)
+    if (strcmp(opts[i], "nocase") == 0 ||
+        strcmp(opts[i], "NC") == 0) {
       flags |= REWRITE_RULE_FLAG_NOCASE;
 
-    else if (strcmp(opts[i], "last") == 0 || strcmp(opts[i], "L") == 0)
+    } else if (strcmp(opts[i], "last") == 0 ||
+               strcmp(opts[i], "L") == 0) {
       flags |= REWRITE_RULE_FLAG_LAST;
+    }
   }
 
   return flags;
@@ -398,54 +439,74 @@ static unsigned char rewrite_match_cond(cmd_rec *cmd, config_rec *cond) {
       rewrite_log("rewrite_match_cond(): checking regex cond");
 
       memset(&rewrite_cond_matches, '\0', sizeof(rewrite_cond_matches));
-      rewrite_cond_matches.match_string = cond->argv[0];
+      rewrite_cond_matches.match_string = cond_str;
       return rewrite_regexec(cond_str, (regex_t *) cond->argv[1], negated,
         &rewrite_cond_matches);
     }
 
     case REWRITE_COND_OP_TEST_DIR: {
+      int res = FALSE;
       struct stat st;
       rewrite_log("rewrite_match_cond(): checking dir test cond");
 
       pr_fs_clear_cache();
-      if (pr_fsio_lstat(cond_str, &st) >= 0 && S_ISDIR(st.st_mode))
-        return TRUE;
+      if (pr_fsio_lstat(cond_str, &st) >= 0 &&
+          S_ISDIR(st.st_mode))
+        res = TRUE;
 
-      return FALSE;
+      if (!negated)
+        return res;
+      else
+        return (res == TRUE ? FALSE : TRUE);
     }
 
     case REWRITE_COND_OP_TEST_FILE: {
+      int res = FALSE;
       struct stat st;
       rewrite_log("rewrite_match_cond(): checking file test cond");
 
       pr_fs_clear_cache();
-      if (pr_fsio_lstat(cond_str, &st) >= 0 && S_ISREG(st.st_mode))
-        return TRUE;
+      if (pr_fsio_lstat(cond_str, &st) >= 0 &&
+          S_ISREG(st.st_mode))
+        res = TRUE;
 
-      return FALSE;
+      if (!negated)
+        return res;
+      else
+        return (res == TRUE ? FALSE : TRUE);
     }
 
     case REWRITE_COND_OP_TEST_SYMLINK: {
+      int res = FALSE;
       struct stat st;
       rewrite_log("rewrite_match_cond(): checking symlink test cond");
 
       pr_fs_clear_cache();
-      if (pr_fsio_lstat(cond_str, &st) >= 0 && S_ISLNK(st.st_mode))
-        return TRUE;
+      if (pr_fsio_lstat(cond_str, &st) >= 0 &&
+          S_ISLNK(st.st_mode))
+        res = TRUE;
 
-      return FALSE;
+      if (!negated)
+        return res;
+      else
+        return (res == TRUE ? FALSE : TRUE);
     }
 
     case REWRITE_COND_OP_TEST_SIZE: {
+      int res = FALSE;
       struct stat st;
       rewrite_log("rewrite_match_cond(): checking size test cond");
 
       pr_fs_clear_cache();
-      if (pr_fsio_lstat(cond_str, &st) >= 0 && S_ISREG(st.st_mode) &&
+      if (pr_fsio_lstat(cond_str, &st) >= 0 &&
+          S_ISREG(st.st_mode) &&
           st.st_size > 0)
-        return TRUE;
+        res = TRUE;
 
-      return FALSE;
+      if (!negated)
+        return res;
+      else
+        return (res == TRUE ? FALSE : TRUE);
     }
 
     default:
@@ -490,7 +551,8 @@ static unsigned char rewrite_parse_map_str(char *str, rewrite_map_t *map) {
     /* OK, now back to our regular schedule parsing... */
     *map_end = '\0';
 
-    if ((tmp = strchr(map_start, ':')) == NULL) {
+    tmp = strchr(map_start, ':');
+    if (tmp == NULL) {
       rewrite_log("rewrite_parse_mapstr(): notice: badly formatted map string");
       return FALSE;
     }
@@ -504,14 +566,16 @@ static unsigned char rewrite_parse_map_str(char *str, rewrite_map_t *map) {
 
     map->map_lookup_key = map_start;
 
-    if ((tmp = strchr(map_start, '|')) != NULL) {
+    tmp = strchr(map_start, '|');
+    if (tmp != NULL) {
       *tmp = '\0';
 
       /* We've got the default value. */
       map->map_default_value = ++tmp;
 
-    } else
+    } else {
       map->map_default_value = "";
+    }
 
     substr = ++map_end;
     return TRUE;
@@ -544,7 +608,8 @@ static unsigned char rewrite_parse_map_txt(rewrite_map_txt_t *txtmap) {
   }
 
   /* Open the file. */
-  if ((ftxt = pr_fsio_open(txtmap->txt_path, O_RDONLY)) == NULL) {
+  ftxt = pr_fsio_open(txtmap->txt_path, O_RDONLY);
+  if (ftxt == NULL) {
     rewrite_log("rewrite_parse_map_txt(): unable to open %s: %s",
       txtmap->txt_path, strerror(errno));
     return FALSE;
@@ -638,7 +703,8 @@ static regex_t *rewrite_regalloc(void) {
   if (!rewrite_regexes)
     rewrite_regexes = make_array(rewrite_pool, 0, sizeof(regex_t *));
 
-  if ((preg = calloc(1, sizeof(regex_t))) == NULL) {
+  preg = calloc(1, sizeof(regex_t));
+  if (preg == NULL) {
     rewrite_log("fatal: memory exhausted during regex_t allocation");
     exit(1);
   }
@@ -683,6 +749,22 @@ static unsigned char rewrite_regexec(const char *string, regex_t *regbuf,
   return have_match;
 }
 
+static void rewrite_replace_cmd_arg(cmd_rec *cmd, char *new_arg) {
+  if (strcmp(cmd->argv[0], C_SITE) != 0) {
+    cmd->arg = new_arg;
+
+  } else {
+    if (strcasecmp(cmd->argv[1], "CHGRP") == 0 ||
+        strcasecmp(cmd->argv[1], "CHMOD") == 0) {
+      cmd->arg = pstrcat(cmd->pool, cmd->argv[1], " ", new_arg, NULL);
+
+    } else {
+      /* Not one of the handled SITE commands. */
+      cmd->arg = new_arg;
+    }
+  }
+}
+
 static char *rewrite_subst(cmd_rec *cmd, char *pattern) {
   char *new_pattern = NULL;
   rewrite_log("rewrite_subst(): original pattern: '%s'", pattern);
@@ -711,43 +793,63 @@ static char *rewrite_subst(cmd_rec *cmd, char *pattern) {
 static char *rewrite_subst_backrefs(cmd_rec *cmd, char *pattern,
     rewrite_match_t *matches) {
   register unsigned int i = 0;
-  char *new_pattern = NULL;
+  char *replacement_pattern = NULL;
 
   for (i = 0; i < REWRITE_MAX_MATCHES; i++) {
+    char buf[3] = {'\0'};
+
+    memset(buf, '\0', sizeof(buf));
+
+    if (matches == &rewrite_rule_matches) {
+      /* Substitute "$N" backreferences for RewriteRule matches */
+      snprintf(buf, sizeof(buf), "$%u", i);
+
+    } else if (matches == &rewrite_cond_matches) {
+      /* Substitute "%N" backreferences for RewriteCondition matches */
+      snprintf(buf, sizeof(buf), "%%%u", i);
+    }
+
+    if (!replacement_pattern)
+      replacement_pattern = pstrdup(cmd->pool, pattern);
+
+    /* Make sure there's a backreference for this in the substitution
+     * pattern.  Otherwise, just continue on.
+     */
+    if (strstr(replacement_pattern, buf) == NULL)
+      continue;
+
     if (matches->match_groups[i].rm_so != -1) {
-      char buf[3] = {'\0'}, tmp;
+      char tmp;
 
-      if (matches == &rewrite_rule_matches)
-        /* Substitute "$N" backreferences for RewriteRule matches */
-        snprintf(buf, sizeof(buf), "$%u", i);
-
-      else if (matches == &rewrite_cond_matches)
-        /* Substitute "%N" backreferences for RewriteCondition matches */
-        snprintf(buf, sizeof(buf), "%%%u", i);
-
-      if (!new_pattern)
-        new_pattern = pstrdup(cmd->pool, pattern);
-
-      /* Make sure there's a backreference for this in the substitution
-       * pattern.  Otherwise, just continue on.
+      /* There's a match for the backref in the string, substitute in
+       * the backreferenced value.
        */
-      if (!strstr(new_pattern, buf))
-        continue;
 
       tmp = (matches->match_string)[matches->match_groups[i].rm_eo];
       (matches->match_string)[matches->match_groups[i].rm_eo] = '\0';
 
       rewrite_log("rewrite_subst_backrefs(): replacing backref '%s' with '%s'",
         buf, &(matches->match_string)[matches->match_groups[i].rm_so]);
-      new_pattern = sreplace(cmd->pool, new_pattern, buf,
+      replacement_pattern = sreplace(cmd->pool, replacement_pattern, buf,
         &(matches->match_string)[matches->match_groups[i].rm_so], NULL);
    
       /* Undo the twiddling of the NUL character. */ 
       (matches->match_string)[matches->match_groups[i].rm_eo] = tmp;
+
+    } else {
+      /* There's backreference in the string, but there no matching
+       * group (i.e. backreferenced value).  Substitute in an empty string
+       * for the backref.
+       */
+
+      rewrite_log("rewrite_subst_backrefs(): replacing backref '%s' with "
+        "empty string", buf);
+      replacement_pattern = sreplace(cmd->pool, replacement_pattern, buf,
+        "", NULL);
     }
   }
 
-  return (new_pattern ? new_pattern : pattern);
+  return (replacement_pattern ? replacement_pattern : pattern);
 }
 
 static char *rewrite_subst_maps(cmd_rec *cmd, char *pattern) {
@@ -774,7 +876,7 @@ static char *rewrite_subst_maps(cmd_rec *cmd, char *pattern) {
 
     while (c) {
 
-      if (!strcmp(c->argv[0], map.map_name)) { 
+      if (strcmp(c->argv[0], map.map_name) == 0) { 
         char *lookup_value = NULL;
         have_map = TRUE;
 
@@ -782,19 +884,19 @@ static char *rewrite_subst_maps(cmd_rec *cmd, char *pattern) {
           map.map_lookup_key, map.map_name);
 
         /* Handle FIFO maps */
-        if (!strcmp(c->argv[1], "fifo")) {
+        if (strcmp(c->argv[1], "fifo") == 0) {
           lookup_value = rewrite_subst_maps_fifo(cmd, c, &map);
           rewrite_log("rewrite_subst_maps(): fifo map '%s' returned '%s'",
             map.map_name, lookup_value);
 
         /* Handle maps of internal functions */
-        } else if (!strcmp(c->argv[1], "int")) {
+        } else if (strcmp(c->argv[1], "int") == 0) {
           lookup_value = rewrite_subst_maps_int(cmd, c, &map);
           rewrite_log("rewrite_subst_maps(): internal map '%s' returned '%s'",
             map.map_name, lookup_value);
 
         /* Handle external file maps */
-        } else if (!strcmp(c->argv[1], "txt")) {
+        } else if (strcmp(c->argv[1], "txt") == 0) {
           lookup_value = rewrite_subst_maps_txt(cmd, c, &map);
           rewrite_log("rewrite_subst_maps(): txt map '%s' returned '%s'",
             map.map_name, lookup_value);
@@ -850,13 +952,15 @@ static char *rewrite_subst_maps_fifo(cmd_rec *cmd, config_rec *c,
   pr_signals_block();
 
   /* See if a RewriteLock has been configured. */
-  if ((fifo_lockname = (char *) get_param_ptr(main_server->conf, "RewriteLock",
-      FALSE)) != NULL) {
-
+  fifo_lockname = (char *) get_param_ptr(main_server->conf, "RewriteLock",
+    FALSE);
+  if (fifo_lockname != NULL) {
     /* Make sure the file exists. */
-    if ((fifo_lockfd = open(fifo_lockname, O_CREAT)) < 0)
+    fifo_lockfd = open(fifo_lockname, O_RDWR|O_CREAT, 0666);
+    if (fifo_lockfd < 0) {
       rewrite_log("rewrite_subst_maps_fifo(): error creating '%s': %s",
         fifo_lockname, strerror(errno));
+    }
   }
 
   /* Obtain a write lock on the lock file, if configured */
@@ -944,8 +1048,8 @@ static char *rewrite_subst_maps_fifo(cmd_rec *cmd, config_rec *c,
 
   /* Read the value from the FIFO, if any. Unblock signals before doing so. */
   pr_signals_unblock();
-  if ((res = rewrite_read_fifo(fifo_fd, value, REWRITE_FIFO_MAXLEN)) <= 0) {
-    
+  res = rewrite_read_fifo(fifo_fd, value, REWRITE_FIFO_MAXLEN);
+  if (res <= 0) {
     if (res < 0)
       rewrite_log("rewrite_subst_maps_fifo(): error reading value from FIFO "
         "'%s': %s", fifo, strerror(errno));
@@ -1033,7 +1137,7 @@ static char *rewrite_subst_maps_txt(cmd_rec *cmd, config_rec *c,
   txt_vals = (char **) txtmap->txt_values;
 
   for (i = 0; i < txtmap->txt_nents; i++)
-    if (!strcmp(txtmap->txt_keys[i], map->map_lookup_key))
+    if (strcmp(txtmap->txt_keys[i], map->map_lookup_key) == 0)
       value = txtmap->txt_values[i];
 
   if (!value)
@@ -1050,10 +1154,11 @@ static char *rewrite_subst_vars(cmd_rec *cmd, char *pattern) {
     char *val = NULL;
 
     /* Does this variable occur in the substitution pattern? */
-    if (!strstr(pattern, rewrite_vars[i]))
+    if (strstr(pattern, rewrite_vars[i]) == NULL)
       continue;
 
-    if ((val = rewrite_expand_var(cmd, pattern, rewrite_vars[i])) != NULL) {
+    val = rewrite_expand_var(cmd, pattern, rewrite_vars[i]);
+    if (val != NULL) {
       rewrite_log("rewrite_subst_vars(): replacing variable '%s' with '%s'",
         rewrite_vars[i], val);
       if (!new_pattern)
@@ -1210,22 +1315,22 @@ static char *rewrite_map_int_replaceall(pool *map_pool, char *key) {
    * the given key, and the sequences to replace for this function.
    */
   char *str = pstrdup(map_pool, key + 1);
-  
+
   tmp = strchr(str , sep);
   if (tmp == NULL) {
-    rewrite_log("rewrite_map_int_replace(): badly formatted input key");
+    rewrite_log("rewrite_map_int_replaceall(): badly formatted input key");
     return NULL;
   }
 
   *tmp = '\0';
   value = str;
-  rewrite_log("rewrite_map_int_replace(): actual key: '%s'", value); 
+  rewrite_log("rewrite_map_int_replaceall(): actual key: '%s'", value); 
  
   str = tmp + 1;
 
   tmp = strchr(str, sep);
   if (tmp == NULL) {
-    rewrite_log("rewrite_map_int_replace(): badly formatted input key");
+    rewrite_log("rewrite_map_int_replaceall(): badly formatted input key");
     return NULL;
   }
 
@@ -1233,12 +1338,12 @@ static char *rewrite_map_int_replaceall(pool *map_pool, char *key) {
   src = str;
   dst = tmp + 1;
   
-  rewrite_log("rewrite_map_int_replace(): replacing '%s' with '%s'", src,
+  rewrite_log("rewrite_map_int_replaceall(): replacing '%s' with '%s'", src,
     dst);
 
   /* Make sure the source sequence is present in the given key. */
   if (strstr(value, src) == NULL) {
-    rewrite_log("rewrite_map_int_replace(): '%s' does not occur in given "
+    rewrite_log("rewrite_map_int_replaceall(): '%s' does not occur in given "
       "key '%s'", src, value);
     return NULL;
   }
@@ -1279,10 +1384,10 @@ static char *rewrite_map_int_unescape(pool *map_pool, char *key) {
   char *value = pcalloc(map_pool, sizeof(char) * strlen(key));
 
   for (i = 0, j = 0; key[j]; ++i, ++j) {
-    if (key[j] != '%')
+    if (key[j] != '%') {
       value[i] = key[j];
 
-    else {
+    } else {
 
       if (!REWRITE_IS_XDIGIT(key[j+1]) || !REWRITE_IS_XDIGIT(key[j+2])) {
         rewrite_log("rewrite_map_int_unescape(): bad escape sequence '%c%c%c'",
@@ -1312,7 +1417,8 @@ static unsigned char rewrite_open_fifo(config_rec *c) {
   /* No interruptions, please. */
   pr_signals_block();
 
-  if ((fd = open(fifo, O_RDWR|O_NONBLOCK)) < 0) {
+  fd = open(fifo, O_RDWR|O_NONBLOCK);
+  if (fd < 0) {
     rewrite_log("rewrite_rdopen_fifo(): unable to open FIFO '%s': %s", fifo,
       strerror(errno));
     pr_signals_unblock();
@@ -1458,13 +1564,13 @@ static void rewrite_openlog(void) {
   if (rewrite_logfd >= 0)
     return;
 
-  if ((rewrite_logfile = get_param_ptr(main_server->conf, "RewriteLog",
-      FALSE)) == NULL) {
+  rewrite_logfile = get_param_ptr(main_server->conf, "RewriteLog", FALSE);
+  if (rewrite_logfile == NULL) {
     rewrite_logfd = -2;
     return;
   }
 
-  if (!strcasecmp(rewrite_logfile, "none")) {
+  if (strcasecmp(rewrite_logfile, "none") == 0) {
     rewrite_logfd = -1;
     rewrite_logfile = NULL;
     return;
@@ -1507,7 +1613,7 @@ static void rewrite_log(char *format, ...) {
   char prefix[80] = {'\0'};
   char entry[1156] = {'\0'};
   time_t logtime = time(NULL);
-  struct tm *timestamp = localtime(&logtime);
+  struct tm *timestamp = pr_localtime(NULL, &logtime);
   va_list mesg;
 
   /* Format the log message */
@@ -1623,7 +1729,8 @@ MODRET set_rewritecondition(cmd_rec *cmd) {
     cond_op = REWRITE_COND_OP_REGEX;
     cond_data = rewrite_regalloc();
 
-    if ((res = regcomp(cond_data, cmd->argv[2], regex_flags)) != 0) {
+    res = regcomp(cond_data, cmd->argv[2], regex_flags);
+    if (res != 0) {
       char errstr[200] = {'\0'};
 
       regerror(res, cond_data, errstr, sizeof(errstr));
@@ -1684,7 +1791,7 @@ MODRET set_rewritecondition(cmd_rec *cmd) {
    */
   *((config_rec **) push_array(rewrite_conds)) = c;
 
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /* usage: RewriteEngine on|off */
@@ -1695,7 +1802,8 @@ MODRET set_rewriteengine(cmd_rec *cmd) {
   CHECK_ARGS(cmd, 1);
   CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL);
 
-  if ((bool = get_boolean(cmd, 1)) == -1)
+  bool = get_boolean(cmd, 1);
+  if (bool == -1)
     CONF_ERROR(cmd, "expecting boolean argument");
 
   /* Check for duplicates */
@@ -1707,7 +1815,7 @@ MODRET set_rewriteengine(cmd_rec *cmd) {
   c->argv[0] = pcalloc(c->pool, sizeof(unsigned char));
   *((unsigned char *) c->argv[0]) = bool;
 
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /* usage: RewriteLock file */
@@ -1721,7 +1829,7 @@ MODRET set_rewritelock(cmd_rec *cmd) {
 
   add_config_param_str(cmd->argv[0], 1, cmd->argv[1]);
 
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /* usage: RewriteLog file|"none" */
@@ -1736,7 +1844,7 @@ MODRET set_rewritelog(cmd_rec *cmd) {
 
   add_config_param_str(cmd->argv[0], 1, cmd->argv[1]);
 
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /* usage: RewriteMap map-name map-type:map-source */ 
@@ -1749,14 +1857,15 @@ MODRET set_rewritemap(cmd_rec *cmd) {
   CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL);
 
   /* Check the configured map types */
-  if ((mapsrc = strchr(cmd->argv[2], ':')) == NULL)
+  mapsrc = strchr(cmd->argv[2], ':');
+  if (mapsrc == NULL)
     CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, "invalid RewriteMap parameter: '",
       cmd->argv[2], "'", NULL));
 
   *mapsrc = '\0';
   mapsrc++;
 
-  if (!strcmp(cmd->argv[2], "int")) {
+  if (strcmp(cmd->argv[2], "int") == 0) {
     c = add_config_param(cmd->argv[0], 3, NULL, NULL, NULL);
 
     /* Check that the given function is a valid internal mapping function. */
@@ -1779,7 +1888,7 @@ MODRET set_rewritemap(cmd_rec *cmd) {
       CONF_ERROR(cmd, pstrcat(cmd->tmp_pool,
         "unknown internal map function requested: '", mapsrc, "'", NULL));
 
-  } else if (!strcmp(cmd->argv[2], "fifo")) {
+  } else if (strcmp(cmd->argv[2], "fifo") == 0) {
     struct stat st;
 
     c = add_config_param(cmd->argv[0], 4, NULL, NULL, NULL, NULL);
@@ -1804,7 +1913,7 @@ MODRET set_rewritemap(cmd_rec *cmd) {
     c->argv[3] = pcalloc(c->pool, sizeof(int));
     *((int *) c->argv[3]) = -1;
 
-  } else if (!strcmp(cmd->argv[2], "txt")) {
+  } else if (strcmp(cmd->argv[2], "txt") == 0) {
     pool *txt_pool = NULL;
     rewrite_map_txt_t *txtmap = NULL;
 
@@ -1840,7 +1949,7 @@ MODRET set_rewritemap(cmd_rec *cmd) {
   c->argv[1] = pstrdup(c->pool, cmd->argv[2]);
   c->argv[2] = map;
 
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /* usage: RewriteRule pattern substitution [flags] */
@@ -1881,7 +1990,8 @@ MODRET set_rewriterule(cmd_rec *cmd) {
     cmd->argv[1]++;
   }
 
-  if ((res = regcomp(preg, cmd->argv[1], regex_flags)) != 0) {
+  res = regcomp(preg, cmd->argv[1], regex_flags);
+  if (res != 0) {
     char errstr[200] = {'\0'};
 
     regerror(res, preg, errstr, sizeof(errstr));
@@ -1941,7 +2051,7 @@ MODRET set_rewriterule(cmd_rec *cmd) {
   *((unsigned int *) c->argv[5]) = rewrite_nrules++;
 
   c->flags |= CF_MERGEDOWN_MULTI;
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /* Command handlers
@@ -1950,17 +2060,45 @@ MODRET set_rewriterule(cmd_rec *cmd) {
 MODRET rewrite_fixup(cmd_rec *cmd) {
   config_rec *c = NULL;
   array_header *seen_rules = NULL;
+  char *cmd_name, *cmd_arg;
 
   /* Is RewriteEngine on? */
   if (!rewrite_engine)
-    return DECLINED(cmd);
+    return PR_DECLINED(cmd);
 
   /* If this command has no argument(s), the module has nothing on which to
    * operate.
    */
   if (cmd->argc == 1) {
     rewrite_log("rewrite_fixup(): skipping %s (no arg)", cmd->argv[0]);
-    return DECLINED(cmd);
+    return PR_DECLINED(cmd);
+  }
+
+  /* If this is a SITE command, handle things a little differently, so that
+   * the rest of the rewrite machinery works properly.
+   */
+  if (strcmp(cmd->argv[0], C_SITE) != 0) {
+    cmd_name = cmd->argv[0];
+    cmd_arg = cmd->arg;
+
+  } else {
+    if (strcasecmp(cmd->argv[1], "CHGRP") == 0 ||
+        strcasecmp(cmd->argv[1], "CHMOD") == 0) {
+      register unsigned int i;
+      char *tmp = "";
+
+      cmd_name = pstrcat(cmd->pool, cmd->argv[0], " ", cmd->argv[1], NULL);
+
+      for (i = 2; i <= cmd->argc-1; i++) {
+        tmp = pstrcat(cmd->pool, tmp, *tmp ? " " : "", cmd->argv[i], NULL);
+      }
+ 
+      cmd_arg = tmp;
+
+    } else {
+      cmd_name = cmd->argv[0];
+      cmd_arg = cmd->arg;
+    }
   }
 
   /* Create an array that will contain the IDs of the RewriteRules we've
@@ -2000,11 +2138,11 @@ MODRET rewrite_fixup(cmd_rec *cmd) {
 
     /* Make sure the given RewriteRule regex matches the command argument. */
     memset(&rewrite_rule_matches, '\0', sizeof(rewrite_rule_matches));
-    rewrite_rule_matches.match_string = cmd->arg;
-    if (!rewrite_regexec(cmd->arg, (regex_t *) c->argv[0],
+    rewrite_rule_matches.match_string = cmd_arg;
+    if (!rewrite_regexec(cmd_arg, (regex_t *) c->argv[0],
         *((unsigned char *) c->argv[2]), &rewrite_rule_matches)) {
       rewrite_log("rewrite_fixup(): %s arg '%s' does not match RewriteRule "
-        "regex", cmd->argv[0], cmd->arg);
+        "regex", cmd_name, cmd_arg);
       c = find_config_next(c, c->next, CONF_PARAM, "RewriteRule", FALSE);
       continue;
 
@@ -2053,8 +2191,44 @@ MODRET rewrite_fixup(cmd_rec *cmd) {
       new_arg = rewrite_subst(cmd, (char *) c->argv[1]);
 
       if (strlen(new_arg) > 0) {
-        cmd->arg = new_arg;
-        rewrite_log("rewrite_fixup(): %s arg now '%s'", cmd->argv[0], cmd->arg);
+        int flags = PR_STR_FL_PRESERVE_COMMENTS;
+        char *param, *dup;
+        array_header *list;
+
+        rewrite_replace_cmd_arg(cmd, new_arg);
+        rewrite_log("rewrite_fixup(): %s arg now '%s'", cmd_name, new_arg);
+
+        /* Be sure to overwrite the entire cmd->argv array, not just
+         * cmd->arg.
+         */
+        cmd->argc = 0;
+        list = make_array(cmd->pool, 2, sizeof(char *));
+
+        *((char **) push_array(list)) = pstrdup(cmd->pool, cmd->argv[0]);
+        cmd->argc++;
+
+        if (strcmp(cmd->argv[0], C_SITE) == 0) {
+          flags |= PR_STR_FL_PRESERVE_WHITESPACE;
+
+          if (strcasecmp(cmd->argv[1], "CHGRP") == 0 ||
+              strcasecmp(cmd->argv[1], "CHMOD") == 0) {
+            *((char **) push_array(list)) = pstrdup(cmd->pool, cmd->argv[1]);
+            cmd->argc++;
+          }
+        }
+
+        dup = pstrdup(cmd->tmp_pool, new_arg);
+        while ((param = pr_str_get_word(&dup, flags)) != NULL) {
+          pr_signals_handle();
+
+          *((char **) push_array(list)) = pstrdup(cmd->pool, param);
+          cmd->argc++;
+        }
+
+        /* NULL-terminate the list. */
+        *((char **) push_array(list)) = NULL;
+
+        cmd->argv = (char **) list->elts;
 
       } else
         rewrite_log("rewrite_fixup(): error processing RewriteRule");
@@ -2067,10 +2241,15 @@ MODRET rewrite_fixup(cmd_rec *cmd) {
       }
     }
 
+    /* When processing multiple RewriteRules, we may have changed cmd->arg.
+     * Thus we need to update the locally cached version of it.
+     */
+    cmd_arg = cmd->arg;
+
     c = find_config_next(c, c->next, CONF_PARAM, "RewriteRule", FALSE);
   }
 
-  return DECLINED(cmd);
+  return PR_DECLINED(cmd);
 }
 
 /* Events handlers
@@ -2149,7 +2328,7 @@ static int rewrite_sess_init(void) {
   while (c) {
     pr_signals_handle();
 
-    if (!strcmp(c->argv[1], "fifo")) {
+    if (strcmp(c->argv[1], "fifo") == 0) {
       PRIVS_ROOT
       if (!rewrite_open_fifo(c))
         rewrite_log("error preparing FIFO RewriteMap");
@@ -2224,7 +2403,10 @@ module rewrite_module = {
   rewrite_init,
 
   /* Session initialization function */
-  rewrite_sess_init
+  rewrite_sess_init,
+
+  /* Module version */
+  MOD_REWRITE_VERSION
 };
 
 #endif /* HAVE_REGEX_H */
